%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This is the main file for the chart generator
%%
%% Author: Aurelien Giraud, aurelien@cl.uni-bremen.de
%%        Universitt Bremen
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% PREREQUISITES
%%
%%  o These can be given in theory.pl:
%%          syntactic_object(<type being the mgsat of all syntactic objects>).
%%
%%          ind_path([path,to,index,of,the,mrs]).
%%
%%          cont_path([path,to,mrs]).
%%          liszt_path([path,to,list,of,EPs]).
%%          hcons_path([path,to,handle,constraints]).
%%
%%     Here is what I use. There are chances that it is similar to that:
%%
%%          syntactic_object(syntactic_object).
%%
%%          ind_path([synsem,loc,cont,ind]).
%%          cont_path([synsem,loc,cont]).
%%          liszt_path([synsem,loc,cont,rels]).
%%          hcons_path([synsem,loc,cont,h_cons]).
%%
%% LOADING
%%
%%  o either launch trale with otion:
%%          -e "compile('/path/to/generator/directory/generator.pl')"
%%    or call within trale:
%%          compile('/path/to/generator/directory/generator.pl').
%%
%% USE
%%
%%  o To parse a sentence and generate from its semantics: 
%%          p_and_g([sentence,as,list,of,words]).
%%
%%  o To parse a phrase of description Desc and generate from its semantics:
%%          p_and_g([sentence,as,list,of,words],Desc).
%%
%%  o To generate from a description Desc (representing the whole sign, not only the
%%    semantics), getting all the outputs in the list Words, and matching Desc to the
%%    description Root:
%%          g(Desc).
%%          g(Desc,Words).
%%          g_d(Desc,Words,Root).
%%          g_d(Desc,Root).
%%
%%  o To display the edges one after the other as soon as they are built 
%%          de_y.
%%    To switch this off:
%%          de_n.
%%
%% TEST SUITE HANDLING
%%
%%  o The test suite handling for generation parses a sentence and tries then to
%%    generate from its semantics.
%%
%%  o Example lines to be added to test_items.pl for generation test suite handling:
%%          tg(6,[pierre,donne,une,fe,,clochette],@root).
%%          tg(6,"Pierre donne une fe  clochette.",@root).
%%
%%  o To test the sentence numbered No in test_items.pl
%%          testg(No).
%%    To test all sentences in test_items.pl
%%          testg(No).
%%  o These can also be used to -write- the test results in file File
%%    (if File exists, its content is replaced by the output)
%%          testgw(No,File).
%%          testgw(all,File).
%%  o These can also be used to -append- the test results to file File
%%    (if File exists, its content is not erased and the results are added at the
%%     end of it)
%%          testga(No,File).
%%          testga(all,File).
%%
%%
%%
%% BUGS:
%% 
%% The bug refered to on 02/02/06 (Indices get mixed sometimes [...]) has been solved.
%%  24/02/06 - Aurelien G.
%% 
%% There is a bug in trale: for some grammars the call to add_to(Moth,Tag,bot) after
%% claise(alec_rule(Name,DtrsDesc,_,Moth,_,_),_true) fails (error or loop)
%% for some grammars. In that case the generator cannot function.
%%  24/02/06 - Aurelien G.
%%
%% Found and corrected a bug in makeAllBits/4. epsBits/1 was not abolished in makeAllBits/4,
%% so when backtracking came back to makeAllBits/4, the assert(epsBits(E)) would add
%% another epsBits predicate without removing the old one. (backtracking to
%% makeAllBits/4 would occur when the input bag of EPs is underspecified)
%%  15/02/06 - Aurelien G.
%%
%% Indices get mixed sometimes in a way that subject and complements of a verb
%% get interchanged leading to more output than it should give.
%%  02/02/06 - Aurelien G.
%% 
%% Does not work yet for difference-lists of EPs in liszt_path(...) and hcons_path(...).
%%  02/02/06 - Aurelien G.



:- use_module(library(system)).
:- dynamic display_edges/1.
:- dynamic blocked_word/1.
:- dynamic blocked_cat/1.
:- dynamic blocked_rule/1.
:- dynamic blocked_phrase/1.
:- dynamic input_bag/1.
:- dynamic rels_length/1.
:- dynamic stopg/0.
:- dynamic slg_chart/1.
:- dynamic slg_chart_copy/1.
:- dynamic allBits/1.
:- dynamic bitPos/2.
:- dynamic c_edge/6.
:- dynamic a_edge/6.
:- dynamic cc_edge/6.
:- dynamic u_no_st_mod/0.
:- dynamic u_st_mod/0.
:- dynamic mod_grRule/3.
:- dynamic non_mod_grRule/3.
:- dynamic u_modRules/1.
:- dynamic u_modCats/1.
:- dynamic gen_words/1.
:- dynamic u_syntactic_object/1.
:- dynamic epsBits/1.
:- dynamic e_er/1.
:- dynamic e_usr/1.
:- dynamic parsed_words_or_desc/1.

%================================================================================
%Debugging / Display predicates
%================================================================================
%--------------------------------------------------------------------------------
%To display the chart and the agenda at their current state with only the words
%and the BitString
%--------------------------------------------------------------------------------
dt:-call_residue(displayThings,_).

displayThings :-
   write('Chart:'),nl,
   dwc,
   write('Agenda:'),nl,
   dwa.

%--------------------------------------------------------------------------------
%To display the agenda at its current state with only the words
%and the BitString
%--------------------------------------------------------------------------------
dwa:-
   (  current_predicate(a_edge,a_edge(_,_,_,_,_,_)))
   -> (dwa2 -> true ; true )
   ;  true.

dwa2:-
   a_edge(_,BS,Ws,_,_,_),
   write(BS),
   write(': '),
   write(Ws),nl,
   fail.

%--------------------------------------------------------------------------------
%To display the chart at its current state with only the words
%and the BitString
%--------------------------------------------------------------------------------
dwc:-
   (  current_predicate(c_edge,c_edge(_,_,_,_,_,_)))
   -> (dwc2 -> true ; true )
   ;  true.

dwc2:-
   c_edge(_,BS,Ws,_,_,_),
   write(BS),write(': '),write(Ws),nl,
   fail.

%--------------------------------------------------------------------------------
%To display the edges one by one when they are added to the agenda or chart
%--------------------------------------------------------------------------------
d_added_edge(A,BS,Ws):-
        write(A),write('   '),
        write(BS),write(': '),write(Ws),nl.

%--------------------------------------------------------------------------------
%
%--------------------------------------------------------------------------------
de_y:-  abolish(display_edges,1),
        assert(display_edges(yes)).
        
de_n:-  abolish(display_edges,1),
        assert(display_edges(no)).

%================================================================================
%General Predicates
%================================================================================
%--------------------------------------------------------------------------------
%open(+L:<list>,-O:<openlist>)
%--------------------------------------------------------------------------------
%O is an open list whose non open part is L.
%--------------------------------------------------------------------------------
open([X],[X|_]).
open([X|Y],[X|Z]):-open(Y,Z).

%--------------------------------------------------------------------------------
%open_h(+L:<list>,-O:<openlist>,-H:<list>)
%--------------------------------------------------------------------------------
%like open/2 but with access to H, the difference between L and O
%--------------------------------------------------------------------------------
open_h([],L,L).
open_h([X|Y],[X|Z],L):-open_h(Y,Z,L).

%--------------------------------------------------------------------------------
%To write a %
%--------------------------------------------------------------------------------
writep(S):-
    name(C,[37]),
    write(C),write(S).

%--------------------------------------------------------------------------------
%To get an independent copy of a FS
%--------------------------------------------------------------------------------
copyFS(FS1,CopyFS1):- copy_term(FS1,CopyFS1).



%================================================================================
%Predicates for handling the list of generated sentences
%================================================================================
%--------------------------------------------------------------------------------
%Initialisation of the list of generated sentences
%--------------------------------------------------------------------------------
initGenWords:-
   abolish(gen_words,1),
   assert(gen_words([])).
   
%--------------------------------------------------------------------------------
%To add a sentence to the list of generated sentences
%--------------------------------------------------------------------------------
add_words(L):-
   gen_words(L1),
   abolish(gen_words,1),
   append(L1,[L],L2),
   assert(gen_words(L2)).



%================================================================================
%Predicates useful for manipulating bitstring bags of EPs
%================================================================================
%--------------------------------------------------------------------------------
%twoPower(+N:<int>, -V:<int>)
%--------------------------------------------------------------------------------
%2^N
%--------------------------------------------------------------------------------
twoPower(0,1).
twoPower(N,V) :- N>0, NMinusOne is N-1, twoPower(NMinusOne,W), V is W*2.

%--------------------------------------------------------------------------------
%getAllBits(+L:<list>, -BitString<int>)
%--------------------------------------------------------------------------------
%get (as an integer) a bitstring AllBits of 1s from a list L
%Allbits and L having the same "length"
%--------------------------------------------------------------------------------
getAllBits(L,AllBits) :-
   getAllBits(L,_,AllBits).

getAllBits([],0,0).
getAllBits([_X|L],N2,V2) :-
   getAllBits(L,N1,V1),
   N2 is N1+1,
   twoPower(N1,W),
   V2 is (V1+W).

%--------------------------------------------------------------------------------
%makeAllBits(+L:<list>, -AllBits:<int>, -EPsBits:<list>)
%--------------------------------------------------------------------------------
%Same as getAllBits plus keeps records of the list elements'bit position
%--------------------------------------------------------------------------------
%What happens if we try to generate from underspecified resl list?
%Shoud I prevent building by backtracking
%so if L is uderspecified, the other vars are underspecified...
%it seems that it would cost too much...

makeAllBits(L,N,AllBits,EPsBits) :-
   makeAllBits2(L,N,AllBits,EPsBits),
   check_liszt_length(N),
   abolish(epsBits,1),
   assert(epsBits(EPsBits)),
   assert(rels_length(N))
   .

makeAllBits2([],0,0,[]).
makeAllBits2([X|L],N2,V2,[bitPos(X,W)|M]) :-
   makeAllBits2(L,N1,V1,M),
   N2 is N1+1,
   twoPower(N1,W),
   V2 is (V1+W).

%--------------------------------------------------------------------------------
%remakeAllBits( +L2:<list>, +AllBitsIn:<int>, +EPsBitsIn:<list>
%--------------------------------------------------------------------------------
%Adds elements to EPsBits
%--------------------------------------------------------------------------------
remakeAllBits(L2,N,AllBits,EPsBits,NewAllBits,NewEPsBits) :-
   remakeAllBits2(L2,N,_,AllBits,NewAllBits,EPsBits2),
   append(EPsBits2,EPsBits,NewEPsBits),
   abolish(epsBits,1),
   assert(epsBits(NewEPsBits)).

remakeAllBits2([],N,N,AllBits,AllBits,[]).

remakeAllBits2([X|L],N,N2,V,V2,[bitPos(X,W)|M]) :-
   remakeAllBits2(L,N,N1,V,V1,M),
   N2 is N1+1,
   twoPower(N1,W),
   V2 is (V1+W).

%--------------------------------------------------------------------------------
%getBitString(+EPsBits:<list>,+EPs:<list>,-N:<int>)
%--------------------------------------------------------------------------------
%It maps a bag of EPs to a bitstring integer.
%It fails if some EPs are present more than once
%and if an EP was not given in the input bag (bitPos/2 would fail).
%--------------------------------------------------------------------------------
getBitString(_EPsBits,[],0).

getBitString(EPsBits,[EP1|EPs],BS) :-
   getBitString(EPsBits,EPs,BS2),
   member(bitPos(EP1,BS1),EPsBits),
   bs_disjoint(BS1,BS2),
   bs_union(BS1,BS2,BS).

%--------------------------------------------------------------------------------
%getMaxBitString(+EPsBits:<list>,-N:<int>)
%--------------------------------------------------------------------------------
%EPsBits is a list of bitpos(EP,BS)
%getMaxBitString/2 gets the bitstring value of a bag which would contain all the
%EPs found in EPsBits
%--------------------------------------------------------------------------------
getMaxBitString(L,EPs,BS) :-
   getMaxBitString(L,EPs,0,BS).

getMaxBitString([],[],BS,BS).
   
getMaxBitString([bitPos(EP1,BS1)|L],[EP1|K],BStemp,BS) :-
   bs_disjoint(BS1,BStemp),
   bs_union(BS1,BStemp,BStemp2),
   getMaxBitString(L,K,BStemp2,BS).
      
%--------------------------------------------------------------------------------
%retrieveEPsBitString(+EPsBits:<list>,-EPs:<list>,-N:<int>)
%--------------------------------------------------------------------------------
%retrieves from a set EPsBits of bitPos(A,B) structures, via backtracking, all
%the possible bags of EPs (As) together with the corresponding bitString value
%of the bag
%--------------------------------------------------------------------------------
retrieveEPsBitString(EPsBits,EPs,BS) :-
   retrieveEPsBitString2(_List1,EPsBits,EPs,0,BS).

retrieveEPsBitString2([],[],[],BS,BS).

retrieveEPsBitString2([bitPos(EP1,BS1)|EPBits1],[bitPos(EP1,BS1)|EPBits],[EP1|L],BStemp,BS) :-
   bs_disjoint(BS1,BStemp),
   bs_union(BS1,BStemp,BStemp2),
   retrieveEPsBitString2(EPBits1,EPBits,L,BStemp2,BS).

retrieveEPsBitString2(L1,[_|EPBits],L,BStemp,BS) :-
   retrieveEPsBitString2(L1,EPBits,L,BStemp,BS).

%--------------------------------------------------------------------------------
%sameBags(+N:<int>,+M:<int>)
%--------------------------------------------------------------------------------
%(bitstring) bag equality
%--------------------------------------------------------------------------------
sameBags(N,N).

%--------------------------------------------------------------------------------
%bs_disjoint(+BitS1:<int>,+BitS2:<int>)
%--------------------------------------------------------------------------------
%(bitstring) bag disjunction
%--------------------------------------------------------------------------------
bs_disjoint(BitS1,BitS2) :-
    epsBits(EPsBits),
    (var(BitS1) -> (member(bitPos(_,BitS1),EPsBits); BitS1=0) ; true),
    (var(BitS2) -> (member(bitPos(_,BitS2),EPsBits); BitS2=0) ; true),
    X is BitS1/\BitS2, X=0.

%--------------------------------------------------------------------------------
%bs_union(+BitS1:<int>,+BitS2:<int>,-NewBitS:<int>)
%--------------------------------------------------------------------------------
%(bitstring) bag union
%--------------------------------------------------------------------------------
bs_union(BitS1,BitS2,NewBitS) :-
    NewBitS is BitS1\/BitS2.


%================================================================================
%Predicates for manipulating edges, chart and agenda
%================================================================================
%--------------------------------------------------------------------------------
%actEdge(+Edge:<edge>)  inactEdge(+Edge:<edge>)
%--------------------------------------------------------------------------------
%Definition/test of active and inactive edges
%--------------------------------------------------------------------------------
actEdge(edge(_Vertex,_Bag_EPs,_Words,_Cat,_Found,[_|_])).

inactEdge(edge(_Vertex,_Bag_EPs,_Words,_Cat,_Found,[])).

%--------------------------------------------------------------------------------
%getEdge(-Edge:<edge>)
%--------------------------------------------------------------------------------
%retrieves an edge from the agenda
%--------------------------------------------------------------------------------
getFromAgenda(edge(A,B,C,D,E,F)) :-
   retract(a_edge(A,B,C,D,E,F)).

%--------------------------------------------------------------------------------
%getFromChart(-Edge:<edge>)
%--------------------------------------------------------------------------------
%retrieves an edge from the chart
%--------------------------------------------------------------------------------
getFromChart(edge(A,B,C,D,E,F)) :-
   retract(c_edge(A,B,C,D,E,F)).

getFromChartCopy(edge(A,B,C,D,E,F)) :-
   retract(cc_edge(A,B,C,D,E,F)).

%--------------------------------------------------------------------------------
%addToChart(+Edge:<edge>)   
%--------------------------------------------------------------------------------
%adds the edge Edge to the chart
%--------------------------------------------------------------------------------
addToChart(edge(A,B,C,D,E,F)) :-
   asserta(c_edge(A,B,C,D,E,F)),
   (display_edges(yes) -> d_added_edge('Chart',B,C) ; true).

%--------------------------------------------------------------------------------
%addToChartCopy(+Edge:<edge>)   
%--------------------------------------------------------------------------------
%adds the edge Edge to the chart copy
%--------------------------------------------------------------------------------
addToChartCopy(edge(A,B,C,D,E,F)) :-
   asserta(cc_edge(A,B,C,D,E,F)).

%--------------------------------------------------------------------------------
%addToAgenda(+Cond:<atom>,+Edge:<edge>)   
%--------------------------------------------------------------------------------
%adds the edge Edge to the Agenda
%Cond=n_b_p used when addToAgenda is called out of the initial building process
%Cond=b_p used when addToAgenda is called from the initial building process
%--------------------------------------------------------------------------------
addToAgenda(Cond,Edge) :-
        addToAgenda2(Cond,Edge)->true;true.

addToAgenda2(_Cond,edge(A,B,C,D,E,F)) :-
        F=[_|_],
        asserta(a_edge(A,B,C,D,E,F)).

addToAgenda2(Cond,edge(A,B,C,D,E,[])) :-
        check_cat(D),        
        asserta(a_edge(A,B,C,D,E,[])),
        (display_edges(yes) -> d_added_edge('Agenda',B,C) ; true),
        (
          Cond=n_b_p
        ->
          (
            epsBits(EPsBits),
            dotMovementCheck(edge(A,B,C,D,E,[]),EPsBits)
          )
        ).
 

%--------------------------------------------------------------------------------
%initChart
%--------------------------------------------------------------------------------
%Chart Initialisation
%--------------------------------------------------------------------------------
initChart :-
   clearChart.

%--------------------------------------------------------------------------------
%Chart deletion
%--------------------------------------------------------------------------------
clearChart :- abolish(c_edge,6).

%--------------------------------------------------------------------------------
%Agenda deletion
%--------------------------------------------------------------------------------
clearAgenda :- abolish(a_edge,6).

%--------------------------------------------------------------------------------
%initAgenda(+EPsBits:<list>)
%--------------------------------------------------------------------------------
%Agenda Initialisation
%--------------------------------------------------------------------------------
initAgenda(EPsBits) :-
   clearAgenda,
   (  (
       initAgenda2(EPsBits))
   -> true
   ;  true).

initAgenda2([]).

initAgenda2(EPsBits):-
   EPsBits=[_|_],
   isAWordOrEmpty(Word,Ref-SVs,W_Bag1,Vert,Kind),
   (Kind=no_u_rels
   ->
       (
           W_Bag1=[]
       
       ->
         (
           e_er(yes),
           fully_deref(Ref,SVs,RefOut,SVsOut),
           addToAgenda(b_p,edge(Vert,0,Word,RefOut-SVsOut,[],[]))
         )
       
       ;
         (
           retrieveEPsBitString(EPsBits,W_Bag2,W_BitStr),
           udListReOrder(W_Bag1,W_Bag2),
           fully_deref(Ref,SVs,RefOut,SVsOut),
           addToAgenda(b_p,edge(Vert,W_BitStr,Word,RefOut-SVsOut,[],[]))
         )
       )
   ;
       (
         e_usr(yes),
         fully_deref(Ref,SVs,RefOut,SVsOut),
         addToAgenda(b_p,edge(Vert,_,Word,RefOut-SVsOut,[],[]))
       )
   ),
   fail.

%--------------------------------------------------------------------------------
%udList(+A:<list>,+B:<list>)
%--------------------------------------------------------------------------------
%unifies the element of two lists from left to right
%--------------------------------------------------------------------------------
udList([A],[B]) :- ud(A,B).

udList([A|L],[B|M]) :-
   ud(A,B),
   udList(L,M).

%--------------------------------------------------------------------------------
%udListReOrder(+A:<list>,+B:<list>,-C:<list>)
%--------------------------------------------------------------------------------
%it does kind of the same job as would be done by calling sameOrderRelsList and
%udList one after the other. Namely it unifies two lists, element by element,
%but in any possible order.
%--------------------------------------------------------------------------------
udListReOrder(L1,L2) :-
   copy_term(L1,L11),
   copy_term(L2,L22),
   length(L11,X),
   length(L22,Y),
   !,
   (  X>=Y
   -> (length(L1,X),
       length(L2,X))
   ;  (length(L1,Y),
       length(L2,Y))),
   udListReOrder2(L1,L2).

udListReOrder2([FS1],[FS2]):- ud_with_chk(FS1,FS2).

udListReOrder2([FS1,FS3|L],L2) :-
  select(FS2,L2,Rest2),
  ud_with_chk(FS1,FS2),
   udListReOrder2([FS3|L],Rest2).

ud_with_chk(T1-A1,T2-A2):-
        ((var(T1),var(A1))
        ->
            (A11=bot,deref(T1,A1,T11,A11))
        ;
            deref(T1,A1,T11,A11)),
        ((var(T2),var(A2))
        ->
            (A22=bot,deref(T2,A2,T22,A22))
        ;
            deref(T2,A2,T22,A22)),
        ud(T11-A11,T22-A22).

%--------------------------------------------------------------------------------
%isAWord(-Word:<list>,-Cat:<FS>,-W_Bag:<list>,-Vert:<FS>)
%--------------------------------------------------------------------------------
%It checks if there is a word in the lexicon and gives its phonology as Word,
%its whole FS as Cat, its bag of EPs as W_Bag, and its vertex/index as Vert.
%It fails if the Word has not a specified 'rels' value
%--------------------------------------------------------------------------------
isAWord(Word,Tag-SVs,Word_Bag,TagOut2-Vert,Kind):- %,TagOut3-Rel) :-
   lex(Word,Tag-SVs),
   check_words([Word]),
   getBagAndVertex(Tag-SVs,Word_Bag,TagOut2-Vert,Kind).

%--------------------------------------------------------------------------------
%isAnEmpty(-Cat:<FS>,-W_Bag:<list>,-Vert:<FS>)
%--------------------------------------------------------------------------------
%like isAWord/4, but for words with an empty phon
%--------------------------------------------------------------------------------
isAnEmpty(Tag-bot,Word_Bag,TagOut2-Vert,Kind):-
        current_predicate(empty,empty(_)),
        empty(Desc),
        add_to(Desc,Tag,bot),
        getBagAndVertex(Tag-bot,Word_Bag,TagOut2-Vert,Kind).

%--------------------------------------------------------------------------------
%isAWordOrEmpty(-Word:<list>,- Cat:<FS>, -W_Bag:<list>, -Vert:<FS>)
%--------------------------------------------------------------------------------
%combines the results of isAWord/4, and isAnEmpty/3
%--------------------------------------------------------------------------------
isAWordOrEmpty(Word,FS,Word_Bag,Vert,Kind):-
        (isAWord(W,FS,Word_Bag,Vert,Kind),
         Word=[W])
        ;
        (isAnEmpty(FS,Word_Bag,Vert,Kind),
         Word=[]).

%--------------------------------------------------------------------------------
%getBagAndVertex(-Cat:<FS>,-W_Bag:<list>,-Vert:<FS>)
%--------------------------------------------------------------------------------
%
%--------------------------------------------------------------------------------
getBagAndVertex(Tag-SVs,Word_Bag,TagOut2-Vert,Kind):-
   check_syntactic_object(Tag-SVs),
   liszt_path(L),
   pathval(L,Tag1,SVs1,TagOut,SVsOut),
   deref(Tag,SVs,Tag1,SVs1),
   (  subsumes_chk(TagOut-SVsOut,_-list)
   -> Kind=u_rels
   ;  (Kind=no_u_rels),
       getListFromFS(TagOut,SVsOut,Word_Bag)),
   ind_path(M),
   pathval(M,Tag2,SVs2,TagOut2,Vert),
   deref(Tag,SVs,Tag2,SVs2).

%--------------------------------------------------------------------------------
%check_syntactic_object(+Cat:<FS>)
%--------------------------------------------------------------------------------
%checks if the word found is not a stem...
%--------------------------------------------------------------------------------
check_syntactic_object(_Tag-SVs):-
   functor(SVs,Type,_),
   u_syntactic_object(SO),
   unify_type(Type,SO,_Res).

%--------------------------------------------------------------------------------
%syntactic_object(+Cat:<FS>)
%--------------------------------------------------------------------------------
%This is made modifyable by the user, if the user does not use this type...
%--------------------------------------------------------------------------------
u_syntactic_object(syntactic_object).

syntactic_object_act:-
   current_predicate(syntactic_object,syntactic_object(_))
   -> syntactic_object(L),
      abolish(u_syntactic_object,1),
      assert(u_syntactic_object(L))
   ;  true.

%--------------------------------------------------------------------------------
%getEPsFromRef(+Tag:<ref>,+SVs:<svs>,-Bag:<list>)
%--------------------------------------------------------------------------------
%Bag is the list of EPs corresponding to the ale-list of EPs found in the
%semantics of the sign Tag-SVs
%--------------------------------------------------------------------------------
getEPsFromRef(Tag,SVs,Bag) :-
   getRelsFromRef(Tag,SVs,Ref2,SVs2),
   assert(input_bag(Ref2-SVs2)),
   getListFromFS(Ref2,SVs2,Bag).

%--------------------------------------------------------------------------------
%getRelsFromRef(+Tag:<ref>,-SVs:<svs>,-Ref2:<ref>,-SVs2:<svs>)
%--------------------------------------------------------------------------------
%Ref2-SVs2 is the liszt part of the sign Tag-SVs
%--------------------------------------------------------------------------------
getRelsFromRef(Tag,SVs,Ref2,SVs2) :-
   liszt_path(L),
   pathval(L,TagOut,SVsOut,Ref2,SVs2),
   deref(Tag,SVs,TagOut,SVsOut).

%--------------------------------------------------------------------------------
%getListFromRef(+Tag:<ref>,+SVs:<svs>,-List:<list>)
%--------------------------------------------------------------------------------
%List is the prolog list corresponding to the ale-list Tag-SVs
%--------------------------------------------------------------------------------
getListFromFS(A,list,_):-var(A),!.
getListFromFS(_A-e_list,list,[]).
getListFromFS(_A,e_list,[]).

getListFromFS(Tag,SVs,[X-SVs1|L]) :-
   pathval([hd],TagOut,SVsOut,X,SVs1),
   pathval([tl],TagOut,SVsOut,Tag2,SVs2),
   deref(Tag,SVs,TagOut,SVsOut),
   getListFromFS(Tag2,SVs2,L).

%--------------------------------------------------------------------------------
%to rebuild the chart from the chart copy
%--------------------------------------------------------------------------------
rebuild_chart:-
   (  rebuild_chart2
   -> true
   ; true).

rebuild_chart2:-
   getFromChartCopy(edge(A,B,C,D,E,F)),
   addToChart(edge(A,B,C,D,E,F)),
   fail.



%================================================================================
%Predicates for grammar rules access and special treatment of modification 
%================================================================================
%--------------------------------------------------------------------------------
%grRule(-RuleName:<atom>,-FS:<fs>,-DtrsList:<list>) :-
%--------------------------------------------------------------------------------
%To retrieve grammar rules from the grammar
%--------------------------------------------------------------------------------
grRule(Name,TagMoth-bot,DtrsList) :-
        clause(alec_rule(Name,DtrsDesc,_,Moth,_,_),true),
        check_rule(Name),
        satisfy_dtrs(DtrsDesc,_DtrCats,[],DtrsList,gdone),
        add_to(Moth,TagMoth,bot).

%-------------------------------------------------------------------------------
%u_modRules(+L:<list>)
%-------------------------------------------------------------------------------
%predicate which may be re-defined by the user via mod_rules/1.
%It says to the generator that no grammar rule perform modification
%--------------------------------------------------------------------------------
u_modRules([]).

%-------------------------------------------------------------------------------
%mod_rules_act(+L:<list>)
%-------------------------------------------------------------------------------
%allows the user to specify the rules performing modification.
%(L is a list of grammar rule names)
%It changes the asserted value of u_modRules/1
%-------------------------------------------------------------------------------
mod_rules_act:-
   current_predicate(mod_rules,mod_rules(_))
   -> mod_rules(L),
      abolish(u_modRules,1),
      assert(u_modRules(L))
   ;  true.

%--------------------------------------------------------------------------------
%isARule(-Moth:<fs>,-DtrsList:<fs>,-RuleName:<atom>)
%--------------------------------------------------------------------------------
%It succeeds if the grammar contains a rule building the FS Moth from the list
%of daughteres DtrsList.
%It only succeeds for non-modification rules when modifiers edges have not
%come back to the agenda. And it only succeeds for modification rules when
%modifiers edges have come back to the agenda.
%--------------------------------------------------------------------------------
isARule(Moth,DtrsList,Name) :-
   (  u_no_st_mod
   -> ( (current_predicate(non_mod_grRule,non_mod_grRule(_,_,_)),
         non_mod_grRule(Name,Moth,DtrsList))
      ;
          (current_predicate(mod_grRule,mod_grRule(_,_,_)),
           mod_grRule(Name,Moth,DtrsList))
      )
   ;
      (
        (  toggleModification(yes)
        -> ( (current_predicate(mod_grRule,mod_grRule(_,_,_)),
              mod_grRule(Name,Moth,DtrsList))
           ; non_mod_grRule(Name,Moth,DtrsList)
           )
        ;  (current_predicate(non_mod_grRule,non_mod_grRule(_,_,_)),
               non_mod_grRule(Name,Moth,DtrsList))
        )
      )
   ).

%--------------------------------------------------------------------------------
%modRule(-RuleName:<atom>)
%--------------------------------------------------------------------------------
%Used to check if a rule is said as performing modification
%--------------------------------------------------------------------------------
modRule(RuleName):-
   u_modRules(L),
   member(RuleName,L).

%--------------------------------------------------------------------------------
%stores the modification and non-modification rules separately as
%[non_]mod_grRule/3  clauses.
%--------------------------------------------------------------------------------
sortRules :-
   abolish(mod_grRule,3),
   abolish(non_mod_grRule,3),
   sortRules1.

sortRules1:-
   ( sortRules2 -> true ; true ).

sortRules2 :-
   grRule(Name,Moth,DtrsList),
   ( (
      modRule(Name))
   ->
     (
      assert(mod_grRule(Name,Moth,DtrsList)) )
   ;
     (
      assert(non_mod_grRule(Name,Moth,DtrsList)) )
   ),
   fail.

%--------------------------------------------------------------------------------
%initialises the modification mechanism:
%stores the modification and non-modification rules separately
%determines the modifiers categories
%--------------------------------------------------------------------------------
initMod :-
   abolish(toggleModification,1),
   assert(toggleModification(no)),
   mod_cats_act,
   mod_rules_act,
   sortRules.


%--------------------------------------------------------------------------------
%specifies the modifiers descriptions
%it would enable to only bring back to the agenda the edges that correspond to a
%potential modifier
%--------------------------------------------------------------------------------
modCat(Tag-SVs):-
   u_modCats(L),     
   member(Type,L),
   pos_path(M),
   pathval(M,TagOut,SVsOut,_Ref2,SVs2),
   deref(Tag,SVs,TagOut,SVsOut),
   functor(SVs2,Type2,_),
   unify_type(Type2,Type,_Res).

mod_cats_act:-
   current_predicate(mod_cats,mod_cats(_))
   -> mod_cats(L),
      abolish(u_modCats,1),
      assert(u_modCats(L))
   ;  true.

u_modCats([]).

%--------------------------------------------------------------------------------
%To switch on or off the special treatment of modification
%--------------------------------------------------------------------------------
u_st_mod(no).

st_mod_act:-
   current_predicate(st_mod,st_mod(_))
   -> st_mod(K),
      abolish(u_st_mod,1),
      assert(u_st_mod(K))
   ;  true.

st_mod:-
   abolish(u_st_mod,1),
   assert(u_st_mod(yes)).

no_st_mod:-
   abolish(u_st_mod,1),
   assert(u_st_mod(no)).

%--------------------------------------------------------------------------------
%When all edges have been processed excluding modification, then
%toggleModification/0 puts modifiers edges back into the agenda and toggles
%modification rules.
%--------------------------------------------------------------------------------
toggleModification :-
   (  u_st_mod(yes)
   -> (  u_modCats([_|_])
      -> (write('Bringing back modification edges to the agenda'),nl,
          toggleModification1,
          rebuild_chart))),
   abolish(toggleModification,1),
   assert(toggleModification(yes)).

toggleModification1 :-
   (  toggleModification2
   -> true
   ;  true).

toggleModification2 :-
     getFromChart(edge(Vert,W_BitStr,Word,Cat,Found,[])), 
     (  modCat(Cat)
     -> addToAgenda(n_b_p,edge(Vert,W_BitStr,Word,Cat,Found,[]))
     ;  addToChartCopy(edge(Vert,W_BitStr,Word,Cat,Found,[])) ),
     fail.



%================================================================================
%Generation predicates
%================================================================================
%--------------------------------------------------------------------------------
%start Symbol
%--------------------------------------------------------------------------------
startSymbol(Tag-SVs):-
      u_gen_root(Root),
      add_to(Root,TagRoot,bot),
      ud(TagRoot-bot,Tag-SVs).

%--------------------------------------------------------------------------------
%to enable the user to specify the categories of the generated FS
%--------------------------------------------------------------------------------
u_gen_root(R):-root_symbol(R).

gen_root_act:-
   current_predicate(gen_root,gen_root(_))
   -> gen_root(Desc), %gen_root might be put in theory.pl by the user to change from default
      abolish(u_gen_root,1),
      assert(u_gen_root(Desc))
   ;  true.

%--------------------------------------------------------------------------------
%g(+Cat:<desc>,-Words:<list>,+Root:<desc>,+Query:<atom>)
%--------------------------------------------------------------------------------
%Words is a list representing the sentence generated from Cat, corresponding to
%the description Root, with Query enabling to switch query_proceed on and off
%--------------------------------------------------------------------------------
g(Cat,Words,Root,Query):-
   call_residue((add_to(Cat,Tag,bot),
                 frozen_term(Tag,Frozen),
                 ((current_predicate(portray_cat,portray_cat(_,_,_,_,_)),
                   portray_cat(_,Cat,Tag,bot,Frozen)) -> true
                 ; nl, write('INITIAL CATEGORY: '), nl, ttyflush,
                     pp_fs_res(Tag,bot,Frozen), nl
                 ),
                 slgGen(Tag,SVs,Words,Root,Query)),Residue),
   ((current_predicate(portray_cat,portray_cat(_,_,_,_,_)),
     portray_cat(Words,Cat,Tag,SVs,Residue)) -> true
   ; nl, write('STRING: '),
     nl, write_list(Words),
     \+ \+ (nl, write('FINAL CATEGORY: '),nl, ttyflush,
            pp_fs_res(Tag,SVs,Residue)), nl
  ).

%--------------------------------------------------------------------------------
%Other versions of g(Cat,Words,Root,Query)
%--------------------------------------------------------------------------------
g(Cat):-
        u_gen_root(Root),
        g(Cat,_Words,Root,query).

g_d(Cat,Root):-
        g(Cat,_Words,Root,query).

g_d(Cat,Words,Root):-
        g(Cat,Words,Root,query).

g(Cat,Words):-
        u_gen_root(Root),
        g(Cat,Words,Root,query).

%--------------------------------------------------------------------------------
%use_root(+R:<desc>)
%--------------------------------------------------------------------------------
%to select to what kind of description the generation output has to correspond to
%--------------------------------------------------------------------------------
use_root(R):-
        abolish(u_gen_root,1),
        assert(u_gen_root(R)).

%--------------------------------------------------------------------------------
%to switch on and off the query_proceed during generation
%(useful for test suites)
%--------------------------------------------------------------------------------
output_query(yes).

no_query_act:-
        abolish(output_query,1),
        assert(output_query(no)).

query_act:-
        abolish(output_query,1),
        assert(output_query(yes)).

%--------------------------------------------------------------------------------
%slgGen(+Tag:<ref>,+SVs:<svs>,-Words:<list>,+Root:<desc>)
%--------------------------------------------------------------------------------
%Words is a phrase generated from Tag-SVs, and corresponding to description Root
%Query enables to switch proceed_query on and off
%--------------------------------------------------------------------------------
slgGen(Tag,SVs,Words,Root,Query) :-
        ( (var(Query) ; Query=query) -> query_act ; no_query_act),
        u_gen_root(R),
        (var(Root) -> true ; use_root(Root)),
        slgGen1(Tag,SVs,Words),
        use_root(R).

%--------------------------------------------------------------------------------
%Other versions of slgGen(Tag,SVs,Words,Root,Query)
%--------------------------------------------------------------------------------
slgGen(Tag,SVs,Words,Root):-
        slgGen(Tag,SVs,Words,Root,query).

slgGen(Tag,SVs,Words) :-
        query_act,
        gen_root_act,
        slgGen1(Tag,SVs,Words).

%--------------------------------------------------------------------------------
%slgGen1(+Tag:<ref>,+SVs:<svs>,-Words:<list>)
%--------------------------------------------------------------------------------
%Words is a sentence or phrase generated from Tag-SVs
%This predicate is called by slgGen predicates (they do some initialisation before)
%--------------------------------------------------------------------------------                
slgGen1(Tag,SVs,Words) :-
   (current_predicate(stopg,stopg) -> abolish(stopg) ; true),
   initGenWords,
   syntactic_object_act,
   st_mod_act,
   write('Initialising the bitstrings...'),nl,
   abolish(epsBits,1),
   abolish(rels_length,1),
   abolish(input_bag,1),
   getEPsFromRef(Tag,SVs,Bag),
   init_bound,
   initMod,
   (
      (   current_predicate(non_mod_grRule,non_mod_grRule(_,_,_))
      ;
         current_predicate(non_mod_grRule,non_mod_grRule(_,_,_)))
   -> true
   ;  write('* no grammar rules have been found'),nl,fail),
   makeAllBits(Bag,_N,AllBits,EPsBits),
   write('Initialising the chart...'),nl,   
   initChart,
   write('Building the agenda...'),nl,
   initAgenda(EPsBits),
   write('Generating new edges...'),nl,
   (
     (
       slgGenProc(Words,AllBits,EPsBits,Tag,SVs)
     )
   ;
     (

       gen_words(Words)
     )
   ).

%--------------------------------------------------------------------------------
%slgGenProc(-S:<list>,+AllBits:<int>,+EPsBits:<list>,+Ref:<ref>,+SVs:<svs>)
%--------------------------------------------------------------------------------
%Processes one by one the edges in the agenda
%--------------------------------------------------------------------------------
slgGenProc(S,AllBits,EPsBits,Ref,SVs) :-
(current_predicate(stopg,stopg) -> fail ; true),
   (
     getFromAgenda(Edge),
     Edge=edge(_,_W_B,_Words,_,_,_)
   )
   ->
   (
     addToChart(Edge),
     (
       slgGenEdgeProc(Edge,AllBits,EPsBits,S,Ref,SVs)
     ;
       slgGenProc(S,AllBits,EPsBits,Ref,SVs)
     )
   )
   ;
   (
     toggleModification(no)
   ->
     toggleModification,
slgGenProc(S,AllBits,EPsBits,Ref,SVs)
   ).


        
%(rels_length(N),M is (N+5),length(Words,D),D<M).
%--------------------------------------------------------------------------------
%slgGenEdgeProc(Edge:<edge>,-S:<list>,+AllBits:<int>,+Words:<list>,+Ref:<ref>,+SVs:<svs>)
%--------------------------------------------------------------------------------
%Generation succeeds if:
%- the edge spans the entire Bag
%- its Cat is a start symbol
%--------------------------------------------------------------------------------
slgGenEdgeProc(edge(_,W_Bag,Words,Tag-SVs,_,[]),AllBits,_EPsBits,Words,_Ref,_SVs2) :-
   sameBags(W_Bag,AllBits),
   startSymbol(Tag-SVs),
   add_words(Words),
   frozen_term([Tag|SVs],Frozen),
   ((current_predicate(portray_cat,portray_cat(_,_,_,_,_)),
     portray_cat(Words,bot,Tag,SVs,Frozen)) -> true
    ; nl, ttyflush,
      pp_fs_res(Tag,SVs,Frozen), nl
   ),
   write(Words),nl,
   (  output_query(yes)
   -> query_proceed, assert(stopg)),
   fail.

%--------------------------------------------------------------------------------
%Invocate Grammar Rules on the inactive edge
%--------------------------------------------------------------------------------
slgGenEdgeProc(edge(V,W_B,Words,Cat,Found,[]),_AllBits,_EPsBits,_S,_,_) :-
   ruleInvocation(edge(V,W_B,Words,Cat,Found,[])).

%--------------------------------------------------------------------------------
%Perform Dot Movement on the active edge
%--------------------------------------------------------------------------------
slgGenEdgeProc(edge(V,W_B,Words,Cat,Found,[X|Cats]),_AllBits,EPsBits,_S,_,_) :-
   dotMovement(edge(V,W_B,Words,Cat,Found,[X|Cats]),EPsBits).

%--------------------------------------------------------------------------------
%removeGoal(-L1:<list>,-L2:<list>).
%--------------------------------------------------------------------------------
%L1 is a list of elements like: cat>_, cats>_, sem_head>_, goal>_
%L2 is the same list, with the goals removed
%--------------------------------------------------------------------------------
removeGoal([],[]).

removeGoal([cat>Cat|Rest],[cat>Cat|Rest2]):-
   removeGoal(Rest,Rest2).

removeGoal([sem_head>Cat|Rest],[sem_head>Cat|Rest2]):-
   removeGoal(Rest,Rest2).

removeGoal([cats>Cat|Rest],[cats>Cat|Rest2]):-
   removeGoal(Rest,Rest2).

removeGoal([goal>_Cat|Rest],Rest2):-
   removeGoal(Rest,Rest2).

%--------------------------------------------------------------------------------
%getVertex(+FS1:<fs>,-FS2:<fs>)
%--------------------------------------------------------------------------------
%FS2 it the vertex of FS1
%--------------------------------------------------------------------------------
getVertex(Tag-SVs,Tag2-SVs2):-
   ind_path(L),
   pathval(L,TagOut,SVsOut,Tag2,SVs2),
   deref(Tag,SVs,TagOut,SVsOut).

%--------------------------------------------------------------------------------
%instanciate_dtrs(+N:<int>,+MRef:<ref>,+MSVs:<svs>,+Cat:<fs>)
%--------------------------------------------------------------------------------
%it unifies the Nth daughter of MRef-MSVs with Cat
%--------------------------------------------------------------------------------
instanciate_dtrs(N,MRef,MSVs,Cat) :-
   dtrs_feat_(Feat),
   pathval([Feat],TagOut,SVsOut,DRef,DSVs),
   deref(MRef,MSVs,TagOut,SVsOut),

   P is N-1,
   length(L,P),
   open_h(L,K,[Cat|_]),
   add_to(K,Tag,bot),
   
   ud(DRef-DSVs,Tag-bot).

instanciate_dtrs(N,MRef-MSVs,Cat) :-
   dtrs_feat_(Feat),
   pathval([Feat],TagOut,SVsOut,DRef,DSVs),
   deref(MRef,MSVs,TagOut,SVsOut),

   Q is N-1,
   length(L,Q),
   open_h(L,K,[Cat|_]),
   add_to(K,Tag,bot),
   
   ud(DRef-DSVs,Tag-bot).

%--------------------------------------------------------------------------------
%instanciate_dtrs_end(+N:<int>,+MRef:<ref>,+MSVs:<svs>)
%--------------------------------------------------------------------------------
%it unifies what could be the N+1th daughter of MRef-MSVs with the e_list, so that
%MRef-MSVs has only N daughters
%--------------------------------------------------------------------------------
instanciate_dtrs_end(N,MRef,MSVs) :-
   dtrs_feat_(Feat),
   pathval([Feat],TagOut,SVsOut,DRef,DSVs),
   deref(MRef,MSVs,TagOut,SVsOut),

   length(L,N),
   add_to(L,Tag,bot),
   
   ud(DRef-DSVs,Tag-bot).

instanciate_dtrs_end(N,MRef-MSVs) :-
   dtrs_feat_(Feat),
   pathval([Feat],TagOut,SVsOut,DRef,DSVs),
   deref(MRef,MSVs,TagOut,SVsOut),

   length(L,N),
   add_to(L,Tag,bot),
   
   ud(DRef-DSVs,Tag-bot).

%--------------------------------------------------------------------------------
%check_args(EPsBits:<list>,Tag-SVs:<fs>,EPsBS:<int>)
%--------------------------------------------------------------------------------
%-
%--------------------------------------------------------------------------------
check_args(EPsBits,Tag-SVs,EPsBS):-
        liszt_path(L),
        pathval(L,Tag2,SVs2,TagOut,SVsOut),
        deref(Tag,SVs,Tag2,SVs2),
        getListFromFS(TagOut,SVsOut,RelsList),
        retrieveEPsBitString(EPsBits,EPsBag,EPsBS),
        copyFS(EPsBag,BaseEPsBag),%!,
        udListReOrder(EPsBag,RelsList),
        subsumes_chk(EPsBag,BaseEPsBag),
        !
                      .

%--------------------------------------------------------------------------------
%Rule Invocation
%--------------------------------------------------------------------------------
%-
%--------------------------------------------------------------------------------
ruleInvocation(edge(_V,W_B,Words,Cat,_Found,[])) :-
   isARule(MRef-MSVs,Cats,_Name),
   removeGoal(Cats,[Kind>Cat1|Rest]),
   (  Rest=[]
   -> getVertex(MRef-MSVs,NewV)
   ; ( Rest=[_Kind2>Cat2|_],
       getVertex(Cat2,NewV))),
   ((Kind=cats) -> copyFS(Cat1,CopyCat1) ; true),
   ud(Cat,Cat1),
   instanciate_dtrs(1,MRef-MSVs,Cat),
   fully_deref(MRef,MSVs,MRefOut,MSVsOut),
   (   Kind=cats
   -> (
        copyFS(MRefOut-MSVsOut,MFS),
        addToAgenda(n_b_p,edge(NewV,W_B,Words,MFS,[Cat],[cats>CopyCat1|Rest])))
   ;   true),
   (   Rest=[]
   -> (
       instanciate_dtrs_end(1,MRefOut-MSVsOut))
   ;   true),
   addToAgenda(n_b_p,edge(NewV,W_B,Words,MRefOut-MSVsOut,[Cat],Rest)),
   fail.

%--------------------------------------------------------------------------------
%Dot Movement
%(more than one Cat left:
%in this case, the new vertex comes from the next Cat needed)
%--------------------------------------------------------------------------------
dotMovement(edge(V,W_B,Words,Cat,Found,[Kind>FS1,Kind2>FS2|Cats]),EPsBits) :-
   (  c_edge(V1,W_B1,Words1,Ref1-SVs1,_Found1,[])
   ;  a_edge(V1,W_B1,Words1,Ref1-SVs1,_Found1,[])),
   ud(V,V1),
   ((Kind=cats) -> trace, copyFS(FS1,CopyFS1) ; true),
   ud(Ref1-SVs1,FS1),
   length(Found,Length),
   N is Length+1,
   bs_disjoint(W_B,W_B1),
   bs_union(W_B,W_B1,NewW_B),
   getVertex(FS2,NewV),
   instanciate_dtrs(N,Cat,Ref1-SVs1),
   append(Words,Words1,NewWords),
   check_words(NewWords),
   fully_deref(Ref1,SVs1,Ref1Out,SVs1Out),
   append(Found,[Ref1Out-SVs1Out],NewFound),
   check_args(EPsBits,Cat,NewW_B),
   addToAgenda(n_b_p,edge(NewV,NewW_B,NewWords,Cat,NewFound,[Kind2>FS2|Cats])),
   (  Kind=cats
   -> (
      getVertex(CopyFS1,NewV2),
      addToAgenda(n_b_p,edge(NewV2,NewW_B,NewWords,Cat,NewFound,[cats>CopyFS1,Kind2>FS2|Cats])))
   ),
   fail.

%--------------------------------------------------------------------------------
%Dot Movement
%(only one Cat left:
%in this case, the vertex comes from the cat of the new inactive edge created)
%--------------------------------------------------------------------------------
dotMovement(edge(V,W_B,Words,Cat,Found,[Kind>FS1]),EPsBits) :-
   (  c_edge(V1,W_B1,Words1,Ref1-SVs1,_Found1,[])
   ;  a_edge(V1,W_B1,Words1,Ref1-SVs1,_Found1,[])),
   ud(V,V1),
   ((Kind=cats) -> trace, copyFS(FS1,CopyFS1) ; true),
   ud(Ref1-SVs1,FS1),
   length(Found,Length),
   N is Length+1,
   bs_disjoint(W_B,W_B1),
   bs_union(W_B,W_B1,NewW_B),
   getVertex(Cat,NewV),
   instanciate_dtrs(N,Cat,Ref1-SVs1),
   append(Words,Words1,NewWords),
   check_words(NewWords),
      fully_deref(Ref1,SVs1,Ref1Out,SVs1Out),
   append(Found,[Ref1Out-SVs1Out],NewFound),
   check_args(EPsBits,Cat,NewW_B),
   (  Kind=cats
   -> (
        getVertex(CopyFS1,NewV2),
        copyFS(Cat,CopyCat),
       addToAgenda(n_b_p,edge(NewV2,NewW_B,NewWords,CopyCat,NewFound,[cats>CopyFS1])))
   ; true
   ),
   instanciate_dtrs_end(N,Cat),
   check_cat(Cat),
   addToAgenda(n_b_p,edge(NewV,NewW_B,NewWords,Cat,NewFound,[])),
   fail.

%--------------------------------------------------------------------------------
%Dot Movement Check
%When a new inactive edge is created, then check dot movement for all old active
%edges (i.e. active edges that are in the chart)
%--------------------------------------------------------------------------------
dotMovementCheck(edge(V1,W_B1,Words1,Ref1-SVs1,_Found1,[]),EPsBits) :-
   c_edge(V,W_B,Words,Cat,Found,[Kind>FS1,Kind2>FS2|Cats]),
   ud(V,V1),
   ((Kind=cats) -> copyFS(FS1,CopyFS1) ; true),     
   ud(FS1,Ref1-SVs1),
   length(Found,Length),
   N is Length+1,
   bs_disjoint(W_B,W_B1),
   bs_union(W_B,W_B1,NewW_B),
   getVertex(FS2,NewV),
   append(Words,Words1,NewWords),
   check_words(NewWords),
   fully_deref(Ref1,SVs1,Ref1Out,SVs1Out),
   append(Found,[Ref1Out-SVs1Out],NewFound),
   instanciate_dtrs(N,Cat,Ref1Out-SVs1Out),
   check_args(EPsBits,Cat,NewW_B),
   addToAgenda(n_b_p,edge(NewV,NewW_B,NewWords,Cat,NewFound,[Kind2>FS2|Cats])),
   (  Kind=cats
   -> (getVertex(CopyFS1,NewV2),
       addToAgenda(n_b_p,edge(NewV2,NewW_B,NewWords,Cat,NewFound,[cats>CopyFS1,Kind2>FS2|Cats])))
   ; true
   ),
   fail.

%--------------------------------------------------------------------------------
%Dot Movement Check
%When a new inactive edge is created, then check dot movement for all old active
%edges (i.e. active edges that are in the chart)
%--------------------------------------------------------------------------------
dotMovementCheck(edge(V1,W_B1,Words1,Ref1-SVs1,_Found1,[]),EPsBits) :-
   c_edge(V,W_B,Words,Cat,Found,[Kind>FS1]),
   ud(V,V1),
   ((Kind=cats) -> copyFS(FS1,CopyFS1) ; true),     
   ud(FS1,Ref1-SVs1),
   length(Found,Length),
   N is Length+1,
   bs_disjoint(W_B,W_B1),
   bs_union(W_B,W_B1,NewW_B),
   getVertex(Cat,NewV),
   append(Words,Words1,NewWords),
   check_words(NewWords),
   fully_deref(Ref1,SVs1,Ref1Out,SVs1Out),
   append(Found,[Ref1Out-SVs1Out],NewFound),
   instanciate_dtrs(N,Cat,Ref1Out-SVs1Out),
   check_args(EPsBits,Cat,NewW_B),
   (  Kind=cats
   -> (getVertex(CopyFS1,NewV2),
       copyFS(Cat,CopyCat),
       addToAgenda(n_b_p,edge(NewV2,NewW_B,NewWords,CopyCat,NewFound,[cats>CopyFS1])))
   ; true
   ),
   instanciate_dtrs_end(N,Cat),
   check_cat(Cat),
   addToAgenda(n_b_p,edge(NewV,NewW_B,NewWords,Cat,NewFound,[])),
   fail.








%================================================================================
%Predicates for test suite handling
%================================================================================
%--------------------------------------------------------------------------------
%to get and write the date and time of a generation in the output file
%--------------------------------------------------------------------------------
write_current_date(Stream):-
        datime(datime(Y,M,D,H,Mi,S)),
        write_date(Stream,Y,M,D,H,Mi,S).

write_date(Stream,Y,M,D,H,Mi,S):-
        write(Stream,D),
        write(Stream,'.'),
        write(Stream,M),
        write(Stream,'.'),
        write(Stream,Y),
        write(Stream,', '),
        write(Stream,H),
        write(Stream,':'),
        write(Stream,Mi),
        write(Stream,':'),
        write(Stream,S).

%--------------------------------------------------------------------------------
%To write a heading in the output file
%--------------------------------------------------------------------------------
heading(File):-
        open(File,write,Stream),
        name(C,[37]),
        write(Stream,C),
        write(Stream,' Output from the test file for generation'),
        nl(Stream),
        close(Stream).

%--------------------------------------------------------------------------------
%To write a small heading with time and date before each item in the output file
%--------------------------------------------------------------------------------
heading2(File):-
        open(File,append,Stream),
        nl(Stream),
        name(C,[37]),
        write(Stream,C),
        write(Stream,' '),
        write_current_date(Stream),
        nl(Stream),
        close(Stream).

%--------------------------------------------------------------------------------
%To remember what words have been parsed and use it in the output file
%--------------------------------------------------------------------------------
parsed_words_or_desc([]).

set_parsed_words_or_desc(L):-
        (
          list_or_string(_Ws_or_Desc,L,_String,list_or_string)
        ;
          list_or_string(L,_WordList,_String,desc)),
        abolish(parsed_words_or_desc,1),
        assert(parsed_words_or_desc(L)).

%--------------------------------------------------------------------------------
%testgr(+N:<int>,+File:<path_to_file>,+Mode:<atom>).
%--------------------------------------------------------------------------------
%Mode takes for value append or write
%This predicate is used to test item number N and write or append the results in
%file File
%--------------------------------------------------------------------------------
testgr(N,File,Mode):-
        (  testg2(N,File,Mode)
        -> true
        ;  true).

%--------------------------------------------------------------------------------
%Other versions of testgr(N,File,Mode)
%--------------------------------------------------------------------------------
testg(all):- testgr(_,no_output_file,_).

testg(N):- testgr(N,no_output_file,_).

testgw(all,File):-
        heading(File),
        testgr(_,File,append).

testgw(N,File):-
        heading(File),
        testgr(N,File,append).

testga(all,File):-                                
        (   file_exists(File)
        ->  true
        ;   heading(File)),
        testgr(_,File,append).

testga(N,File):-
        (file_exists(File)
        ->  true
        ;   heading(File)),
        testgr(N,File,append).

%--------------------------------------------------------------------------------
%used by testgr and other versions
%--------------------------------------------------------------------------------
testg2(N,File,Mode):-
        call_residue(
        (
         tg(N,Ws,R),
         set_parsed_words_or_desc(Ws),
         p_and_g_no_query(Ws,N,R),
         gen_words(L),
         parsed_words_or_desc(L2),
         (  File=no_output_file
         -> true
         ;  heading2(File),
            open(File,Mode,Stream),
            write(Stream,N),
            write(Stream,': '),
            write(Stream,L2),
            nl(Stream),
            write(Stream,L),
            nl(Stream),
            close(Stream)
         ),
         fail
        ) ,_R).

%--------------------------------------------------------------------------------
%
%--------------------------------------------------------------------------------
%Code taken and adapted from file test_suite_handling.pl from Detmar Meurers
%--------------------------------------------------------------------------------
list_or_string(Ws_or_Desc,WordList,String,Kind):-
        (
          [H|_] = Ws_or_Desc
        )
        ->
        (
          integer(H)
        ->  
          Kind=list_or_string,
          Ws_or_Desc=String,
          general_tokenize_sentence_string(String,WordList,_Desc)
        ;                     
          Kind=list_or_string,
          Ws_or_Desc=WordList,
          wordlist2string(WordList,String)
        )
        ;
        Kind = desc.

%--------------------------------------------------------------------------------
%p_and_g_q(+Query:<atom>,Ws_or_Desc:<list or desc>,+N:<int>,+R:<desc>)
%--------------------------------------------------------------------------------
%Either generate from description Ws_or_desc or parse it and generate from the
%parse if it is a phrase.
%Query enables to switch query_proceed on and off
%R is a description with whom generated objects have to match
%--------------------------------------------------------------------------------
p_and_g_q(Query,Ws_or_Desc,N,R):-
        call_residue(
                     (
                       (
                         (
                           list_or_string(Ws_or_Desc,WordList,_String,list_or_string)
                         ->                             
                           init_parse_count,
                           write('* parsing'),(N>=0 -> (write(' '),write(N));true),write(': '),write(WordList),nl,
                           rec(WordList,Tag,SVs,R,_Residue),
                           parse_count_incr,
                           cont_path(Cont),
                           pathval(Cont,Tag,SVs,CTag,CSVs),
                           pathval(Cont,TagOut,bot,CTag2,CSVs2),
                           ud(CTag,CSVs,CTag2,CSVs2),
                           parse_count(M),
                           write('* generating from parse '),write(M),write('...'),nl,
                           slgGen(TagOut,bot,_Words,R,Query)
                         ;  
                           write('* generating from Desc=('),
                           write(Ws_or_Desc),write(') ...'),nl,
                           g(Ws_or_Desc,_Words,R,Query)
                         ),
                         fail)
                     ; true
                     ),_).

%--------------------------------------------------------------------------------
%Other versions of p_and_g_q(Query,Ws_or_Desc,N,R)
%--------------------------------------------------------------------------------
p_and_g_no_query(Ws,N):-
        u_gen_root(R),
        p_and_g_q(no_query,Ws,N,R).

p_and_g_no_query(Ws,N,R):-
        p_and_g_q(no_query,Ws,N,R).

p_and_g(Ws):-
        u_gen_root(R),
        p_and_g_q(query,Ws,(-1),R).

p_and_g(Ws,R):-
        u_gen_root(Root),
        use_root(R),
        p_and_g_q(query,Ws,(-1),R),
        use_root(Root).

%--------------------------------------------------------------------------------
%To count the number of parses found
%--------------------------------------------------------------------------------
parse_count(0).

init_parse_count:-
        abolish(parse_count,1),
        assert(parse_count(0)).

parse_count_incr:-
        parse_count(N),
        abolish(parse_count,1),
        M is N+1,
        assert(parse_count(M)).



%================================================================================
%'Bounding' predicates
%================================================================================

%--------------------------------------------------------------------------------
%not_input_no_words_act
%--------------------------------------------------------------------------------
%I could do all the add_to at grammar compile time...
%
%This predicate enables the user to specify blocking conditions on the edges to
%be built: it checks for the user defined predicate
%not_input_no_words(BagEPsDesc,Words) and if for all EPs X in BagEPsDesc, X
%cannot be unified with any EP in the input bag of EPs, then the edges whose words
%is a member of Words will not be built.
%--------------------------------------------------------------------------------
not_input_no_words_act:-
        current_predicate(not_input_no_words,not_input_no_words(_,_))
        ->
        (
          not_input_no_words_act3
        ->
          true
        ;
          true
        )
        ;
        true.

not_input_no_words_act3:-
        not_input_no_words(BagEPsDesc,Words),
        (
          (var(BagEPsDesc);BagEPsDesc=[])
        -> 
          fail
        ;
          not_in_input(BagEPsDesc),
          no_word_act(Words)
        ),
        fail.

%--------------------------------------------------------------------------------
%not_input_no_cats_act
%--------------------------------------------------------------------------------
%I could do all the add_to at grammar compile time...
%
%This predicate enables the user to specify blocking conditions on the edges to
%be built: it checks for the user defined predicate
%not_input_no_rules(BagEPsDesc,Cats) and if for all EPs X in BagEPsDesc, X
%cannot be unified with any EP in the input bag of EPs, then the edges whose category
%is a member of Cats will not be built.
%--------------------------------------------------------------------------------
not_input_no_cats_act:-
        current_predicate(not_input_no_cats,not_input_no_cats(_,_))
        ->
        (
          not_input_no_cats_act2
        ->
          true
        ;
          true
        )
        ;
        true.

not_input_no_cats_act2:-
        not_input_no_cats(BagEPsDesc,Cats),
        (
          (var(BagEPsDesc);BagEPsDesc=[])
        -> 
          fail
        ;
          not_in_input(BagEPsDesc),
          no_cat_act(Cats)
        ),
        fail.



%--------------------------------------------------------------------------------
%not_input_no_rules_act
%--------------------------------------------------------------------------------
%I could do all the add_to at grammar compile time...
%
%This predicate enables the user to specify blocking conditions on the edges to
%be built: it checks for the user defined predicate
%not_input_no_rules(BagEPsDesc,Names) and if for all EPs X in BagEPsDesc, X
%cannot be unified with any EP in the input bag of EPs, then the rules whose name
%is a member of Names will not be used by the generator.
%--------------------------------------------------------------------------------
not_input_no_rules_act:-
        current_predicate(not_input_no_rules,not_input_no_rules(_,_))
        ->
        (
          not_input_no_rules_act2
        ->
          true
        ;
          true
        )
        ;
        true.

not_input_no_rules_act2:-
        not_input_no_rules(BagEPsDesc,Names),
        (
          (var(BagEPsDesc);BagEPsDesc=[])
        
        -> 
          fail
        ;
          not_in_input(BagEPsDesc),
          no_rules_act(Names)
        ),
        fail.


%--------------------------------------------------------------------------------
%not_in_input(+BagEPsDesc:<list of desc>)
%--------------------------------------------------------------------------------
not_in_input(BagEPsDesc):-
        input_bag(InputBag),
        in_input(BagEPsDesc,InputBag)
        ->
        fail
        ;
        true.

in_input([EP|_],InputBag):-
        add_to(EP,Tag,bot),
        member(IBEP,InputBag),
        ud(Tag-bot,IBEP),
        !.

in_input([_|L],InputBag):-
        in_input(L,InputBag).

%--------------------------------------------------------------------------------
%no_word_act(+Words:<list of lists>)
%--------------------------------------------------------------------------------
%This predicate asserts blocked words and phrases that will be thrown away by
%the generator
%--------------------------------------------------------------------------------
no_word_act([[Word]]):-
        assert(blocked_word([Word])).

no_word_act([[Word1,Word2|Rest]]):-
        assert(blocked_phrase([Word1,Word2|Rest])).

no_word_act([Words1,Words2|Rest]):-
        no_word_act([Words1]),
        no_word_act([Words2|Rest]).

%--------------------------------------------------------------------------------
%no_cat_act(+Cats:<list of descs>)
%--------------------------------------------------------------------------------
%This predicate asserts blocked cats that will be thrown away by the generator
%--------------------------------------------------------------------------------
no_cat_act([Cat]):-
        add_to(Cat,Tag,bot),
        assert(blocked_cat(Tag-bot)).

no_cat_act([Cat1,Cat2|Rest]):-
        no_cat_act([Cat1]),
        no_cat_act([Cat2|Rest]).

%--------------------------------------------------------------------------------
%no_rule_act(+Names:<list>)
%--------------------------------------------------------------------------------
%This predicate asserts names of blocked rules that will be thrown away by
%the generator
%--------------------------------------------------------------------------------
no_rule_act([Name]):-
        assert(blocked_rule(Name)).

no_rule_act([Name1,Name2|Rest]):-
        no_rule_act([Name1]),
        no_rule_act([Name2|Rest]).

%--------------------------------------------------------------------------------
%
%--------------------------------------------------------------------------------
%This predicate enables the user to specify blocking conditions on the edges to
%be built: the edges whose words is a member of Words in the predicate
%ignore_phrases_words(+Words:<list of lists>) will not be built.
%--------------------------------------------------------------------------------
ignore_phrases_words_act:-
        current_predicate(ignore_phrases_words,ignore_phrases_words(_))
        ->
        ignore_phrases_words(L),
        no_word_act(L)
        ;
        true.

%--------------------------------------------------------------------------------
%
%--------------------------------------------------------------------------------
%This predicate enables the user to specify blocking conditions on the edges to
%be built: the edges whose category is a member of Cats in the predicate
%ignore_cats(+Cats:<list of descs>) will not be built.
%--------------------------------------------------------------------------------
ignore_cats_act:-
        current_predicate(ignore_cats,ignore_cats(_))
        ->
        ignore_cats(L),
        no_cat_act(L)
        ;
        true.

%--------------------------------------------------------------------------------
%
%--------------------------------------------------------------------------------
%This predicate enables the user to specify blocking conditions on the rules not
%to be used. The rules whose name is a member of Names in ignore_rules(Names)
%will not be used by the generator
%--------------------------------------------------------------------------------
ignore_rules_act:-
        current_predicate(ignore_rules,ignore_rules(_))
        ->
        ignore_rules(L),
        no_rule_act(L)
        ;
        true.

%--------------------------------------------------------------------------------
%check_words(+Words:<list>)
%--------------------------------------------------------------------------------
check_words(Words):-
        (
          (
            current_predicate(blocked_word,blocked_word(_)),
            blocked_word(Words)
          )
          ;
          (
            current_predicate(blocked_phrase,blocked_phrase(_)),
            blocked_phrase(Words)
          )
        )
        -> fail
        ;
        length(Words,N),
        check_phrase_length(N).
        
%--------------------------------------------------------------------------------
%check_cat(+Cat:<fs>)
%--------------------------------------------------------------------------------
%I use it only before an edge becomes inactive, when it is fully instanciated...
%--------------------------------------------------------------------------------
check_cat(Cat):-
        (
          current_predicate(blocked_cat,blocked_cat(_)),
          blocked_cat(Cat2),
          ud(Cat,Cat2)
        )
        -> fail
        ;
        true.

%--------------------------------------------------------------------------------
%check_rule(+Rule:<fs>)
%--------------------------------------------------------------------------------
check_rule(Name):-
        (
          current_predicate(blocked_rule,blocked_rule(_)),
          blocked_rule(Name)
        )
        -> fail
        ;
        true.

%--------------------------------------------------------------------------------
%check_liszt_length(+N:<int>)
%--------------------------------------------------------------------------------
%if the user has given an underspecified list of rels as input to the genrator
%but specified a maximal length for it, we prevent EPsBits to be longer.
%--------------------------------------------------------------------------------
check_liszt_length(N):-
        current_predicate(max_liszt_length,max_liszt_length(_))
        ->
        max_liszt_length(NN),
        (NN<N -> !,fail ; true)
        ;
        true.

%--------------------------------------------------------------------------------
%check_phrase_length(+N:<int>)
%--------------------------------------------------------------------------------
%if the user has specified a maximal length for the generated phrases,
%we prevent any built phrase to be longer.
%--------------------------------------------------------------------------------
check_phrase_length(N):-
        (
          current_predicate(max_phrase_length,max_phrase_length(_))
        ->
          max_phrase_length(NN),
          (NN<N -> !,fail ; true)
        ;
          true
        ).

%--------------------------------------------------------------------------------
%to initialise the predicates used to bound the search space
%--------------------------------------------------------------------------------
init_bound:-
        abolish(blocked_phrase,1),
        abolish(blocked_word,1),
        abolish(blocked_cat,1),
        ignore_phrases_words_act,
        ignore_cats_act,
        ignore_rules_act,
        not_input_no_words_act,
        not_input_no_rules_act,
        not_input_no_cats_act.

%--------------------------------------------------------------------------------
%to reset the predicates used to bound the search space
%--------------------------------------------------------------------------------
reset_bound:-
        abolish(max_phrase_length,1),
        abolish(max_liszt_length,1),
        abolish(ignore_phrases_words,1),
        abolish(ignore_cats,1),
        abolish(ignore_rules,1),
        abolish(not_input_no_words,1),
        abolish(not_input_no_rules,1),
        abolish(not_input_no_cats,1).

%--------------------------------------------------------------------------------
%ignore edges with unspecified rels (ie_usr) or use them (ue_usr)
%--------------------------------------------------------------------------------
ie_usr:-abolish(e_usr,1),assert(e_usr(no)).
ue_usr:-abolish(e_usr,1),assert(e_usr(yes)).

e_usr(yes).

%--------------------------------------------------------------------------------
%ignore edges whose rels is an empty list (ie_er) or use them (ue_er)
%--------------------------------------------------------------------------------
ie_er:-abolish(e_er,1),assert(e_er(no)).
ue_er:-abolish(e_er,1),assert(e_er(yes)).

e_er(yes).

